@@ -1,6 +1,6 @@
 /*  vcfnorm.c -- Left-align and normalize indels.
 
-    Copyright (C) 2013-2021 Genome Research Ltd.
+    Copyright (C) 2013-2022 Genome Research Ltd.
 
     Author: Petr Danecek <pd3@sanger.ac.uk>
 
@@ -97,12 +97,12 @@ typedef struct
     faidx_t *fai;
     struct { int tot, set, swap; } nref;
     char **argv, *output_fname, *ref_fname, *vcf_fname, *region, *targets;
-    int argc, rmdup, output_type, n_threads, check_ref, strict_filter, do_indels;
+    int argc, rmdup, output_type, n_threads, check_ref, strict_filter, do_indels, clevel;
     int nchanged, nskipped, nsplit, ntotal, mrows_op, mrows_collapse, parsimonious;
     int record_cmd_line, force, force_warned, keep_sum_ad;
     abuf_t *abuf;
     abuf_opt_t atomize;
-    int use_star_allele;
+    int use_star_allele, ma_use_ref_allele;
     char *old_rec_tag;
     htsFile *out;
 }
@@ -142,6 +142,7 @@ static void seq_to_upper(char *seq, int len)
 
 static void fix_ref(args_t *args, bcf1_t *line)
 {
+    bcf_unpack(line, BCF_UN_STR);
     int reflen = strlen(line->d.allele[0]);
     int i,j, maxlen = reflen, len;
     for (i=1; i<line->n_allele; i++)
@@ -160,10 +161,10 @@ static void fix_ref(args_t *args, bcf1_t *line)
     if ( !strncasecmp(line->d.allele[0],ref,reflen) ) { free(ref); return; }
 
     // is the REF allele missing?
-    if ( reflen==1 && line->d.allele[0][0]=='.' ) 
-    { 
-        line->d.allele[0][0] = ref[0]; 
-        args->nref.set++; 
+    if ( reflen==1 && line->d.allele[0][0]=='.' )
+    {
+        line->d.allele[0][0] = ref[0];
+        args->nref.set++;
         free(ref);
         bcf_update_alleles(args->out_hdr,line,(const char**)line->d.allele,line->n_allele);
         return;
@@ -235,7 +236,7 @@ static void fix_ref(args_t *args, bcf1_t *line)
     for (j=1; j<line->n_allele; j++)
     {
         kputc(',',&str);
-        if ( j==i ) 
+        if ( j==i )
             kputs(line->d.allele[0],&str);
         else
             kputs(line->d.allele[j],&str);
@@ -710,11 +711,14 @@ static void split_format_genotype(args_t *args, bcf1_t *src, bcf_fmt_t *fmt, int
         for (j=0; j<ngts; j++)
         {
             if ( gt[j]==bcf_int32_vector_end ) break;
-            if ( bcf_gt_is_missing(gt[j]) || bcf_gt_allele(gt[j])==0 ) continue; // missing allele or ref: leave as is
+            if ( bcf_gt_is_missing(gt[j]) ) continue; // missing allele: leave as is
+            if ( (ialt==0 || args->ma_use_ref_allele) && bcf_gt_allele(gt[j])==0 ) continue; // ref && `--multi-overlaps 0`: leave as is
             if ( bcf_gt_allele(gt[j])==ialt+1 )
                 gt[j] = bcf_gt_unphased(1) | bcf_gt_is_phased(gt[j]); // set to first ALT
-            else
+            else if ( args->ma_use_ref_allele )
                 gt[j] = bcf_gt_unphased(0) | bcf_gt_is_phased(gt[j]); // set to REF
+            else
+                gt[j] = bcf_gt_missing | bcf_gt_is_phased(gt[j]);     // set to missing
         }
         gt += ngts;
     }
@@ -722,7 +726,7 @@ static void split_format_genotype(args_t *args, bcf1_t *src, bcf_fmt_t *fmt, int
 }
 static void split_format_numeric(args_t *args, bcf1_t *src, bcf_fmt_t *fmt, int ialt, bcf1_t *dst)
 {
-    #define BRANCH_NUMERIC(type,type_t,is_vector_end,is_missing,set_vector_end) \
+    #define BRANCH_NUMERIC(type,type_t,is_vector_end,is_missing,set_vector_end,set_missing) \
     { \
         const char *tag = bcf_hdr_int2id(args->hdr,BCF_DT_ID,fmt->id); \
         int ntmp = args->ntmp_arr1 / sizeof(type_t); \
@@ -761,7 +765,10 @@ static void split_format_numeric(args_t *args, bcf1_t *src, bcf_fmt_t *fmt, int
             type_t *src_vals = vals, *dst_vals = vals; \
             for (i=0; i<nsmpl; i++) \
             { \
-                dst_vals[0] = src_vals[ialt]; \
+                int idst = 0; \
+                int isrc = ialt; \
+                if ( is_missing || is_vector_end ) set_missing; \
+                else dst_vals[idst] = src_vals[isrc]; \
                 dst_vals += 1; \
                 src_vals += nvals; \
             } \
@@ -795,8 +802,14 @@ static void split_format_numeric(args_t *args, bcf1_t *src, bcf_fmt_t *fmt, int
                 { \
                     dst_vals[0] = src_vals[0]; \
                     for (j=1; j<nvals; j++) \
+                    { \
+                        int isrc = j; \
                         if ( j!=ialt+1 && !(is_missing) && !(is_vector_end) ) dst_vals[0] += src_vals[j]; \
-                    dst_vals[1] = src_vals[ialt+1]; \
+                    } \
+                    int isrc = ialt + 1; \
+                    int idst = 1; \
+                    if ( is_vector_end ) set_missing; \
+                    else dst_vals[idst] = src_vals[isrc]; \
                     dst_vals += 2; \
                     src_vals += nvals; \
                 } \
@@ -806,7 +819,10 @@ static void split_format_numeric(args_t *args, bcf1_t *src, bcf_fmt_t *fmt, int
                 for (i=0; i<nsmpl; i++) \
                 { \
                     dst_vals[0] = src_vals[0]; \
-                    dst_vals[1] = src_vals[ialt+1]; \
+                    int isrc = ialt + 1; \
+                    int idst = 1; \
+                    if ( is_vector_end ) set_missing; \
+                    else dst_vals[idst] = src_vals[isrc]; \
                     dst_vals += 2; \
                     src_vals += nvals; \
                 } \
@@ -841,14 +857,18 @@ static void split_format_numeric(args_t *args, bcf1_t *src, bcf_fmt_t *fmt, int
                 if ( !haploid ) \
                 { \
                     int j; \
-                    for (j=0; j<nvals; j++) if ( is_vector_end ) break; \
+                    for (j=0; j<nvals; j++) \
+                    { \
+                        int isrc = j; \
+                        if ( is_vector_end ) break; \
+                    } \
                     if ( j!=nvals ) haploid = 1; \
                 } \
                 dst_vals[0] = src_vals[0]; \
                 if ( haploid ) \
                 { \
                     dst_vals[1] = src_vals[ialt+1]; \
-                    if ( !all_haploid ) set_vector_end; \
+                    if ( !all_haploid ) { int idst = 2; set_vector_end; } \
                 } \
                 else \
                 { \
@@ -865,8 +885,8 @@ static void split_format_numeric(args_t *args, bcf1_t *src, bcf_fmt_t *fmt, int
     }
     switch (bcf_hdr_id2type(args->hdr,BCF_HL_FMT,fmt->id))
     {
-        case BCF_HT_INT:  BRANCH_NUMERIC(int32, int32_t, src_vals[j]==bcf_int32_vector_end, src_vals[j]==bcf_int32_missing, dst_vals[2]=bcf_int32_vector_end); break;
-        case BCF_HT_REAL: BRANCH_NUMERIC(float, float, bcf_float_is_vector_end(src_vals[j]), bcf_float_is_missing(src_vals[j]), bcf_float_set_vector_end(dst_vals[2])); break;
+        case BCF_HT_INT:  BRANCH_NUMERIC(int32, int32_t, src_vals[isrc]==bcf_int32_vector_end, src_vals[isrc]==bcf_int32_missing, dst_vals[idst]=bcf_int32_vector_end, dst_vals[idst]=bcf_int32_missing); break;
+        case BCF_HT_REAL: BRANCH_NUMERIC(float, float, bcf_float_is_vector_end(src_vals[isrc]), bcf_float_is_missing(src_vals[isrc]), bcf_float_set_vector_end(dst_vals[idst]), bcf_float_set_missing(src_vals[idst])); break;
     }
     #undef BRANCH_NUMERIC
 }
@@ -1252,7 +1272,7 @@ static void merge_format_genotype(args_t *args, bcf1_t **lines, int nlines, bcf_
     int nsmpl = bcf_hdr_nsamples(args->hdr);
     ngts /= nsmpl;
 
-    int i, j, k;
+    int i, j, k,k2;
     for (i=1; i<nlines; i++)
     {
         int ntmp2 = args->ntmp_arr2 / 4;
@@ -1265,16 +1285,19 @@ static void merge_format_genotype(args_t *args, bcf1_t **lines, int nlines, bcf_
         int32_t *gt2 = (int32_t*) args->tmp_arr2;
         for (j=0; j<nsmpl; j++)
         {
-            for (k=0; k<ngts; k++)
+            for (k2=0; k2<ngts2; k2++)
             {
-                if ( gt2[k]==bcf_int32_vector_end ) break;
-                if ( bcf_gt_is_missing(gt2[k]) || bcf_gt_allele(gt2[k])==0 ) continue;
-                if ( gt2[k]==0 ) gt[k] = 0; // missing genotype
-                else
+                if ( gt2[k2]==bcf_int32_vector_end ) break;
+                if ( bcf_gt_is_missing(gt2[k2]) ) continue;
+                int ial2 = bcf_gt_allele(gt2[k2]);
+                if ( ial2==0 ) continue;    // never overwrite with ref
+                if ( ial2>=args->maps[i].nals ) error("Error at %s:%"PRId64": incorrect allele index %d\n",bcf_seqname(args->hdr,lines[i]),(int64_t) lines[i]->pos+1,ial2);
+                int ial = args->maps[i].map[ial2];
+                for (k=0; k<ngts; k++)
+                    if ( gt[k]==bcf_int32_vector_end || bcf_gt_is_missing(gt[k]) || !bcf_gt_allele(gt[k]) ) break;
+                if ( k<ngts )
                 {
-                    int ial = bcf_gt_allele(gt2[k]);
-                    if ( ial>=args->maps[i].nals ) error("Error at %s:%"PRId64": incorrect allele index %d\n",bcf_seqname(args->hdr,lines[i]),(int64_t) lines[i]->pos+1,ial);
-                    gt[k] = bcf_gt_unphased( args->maps[i].map[ial] ) | bcf_gt_is_phased(gt[k]);
+                    gt[k] = bcf_gt_unphased(ial);
                 }
             }
             gt  += ngts;
@@ -1771,7 +1794,7 @@ static int cmpals_match(cmpals_t *ca, bcf1_t *rec)
         }
 
         khash_t(str2int) *hash = (khash_t(str2int)*) cmpals->hash;
-        for (j=1; j<rec->n_allele; j++) 
+        for (j=1; j<rec->n_allele; j++)
             if ( !khash_str2int_has_key(hash, rec->d.allele[j]) ) break;
         if ( j<rec->n_allele ) continue;
         return 1;
@@ -1860,7 +1883,7 @@ static void init_data(args_t *args)
 
     args->out_hdr = bcf_hdr_dup(args->hdr);
     if ( args->old_rec_tag )
-        bcf_hdr_printf(args->out_hdr,"##INFO=<ID=%s,Number=1,Type=String,Description=\"Original variant. Format: CHR|POS|REF|ALT|USED_ALT_IDX\">",args->old_rec_tag); 
+        bcf_hdr_printf(args->out_hdr,"##INFO=<ID=%s,Number=1,Type=String,Description=\"Original variant. Format: CHR|POS|REF|ALT|USED_ALT_IDX\">",args->old_rec_tag);
 
     rbuf_init(&args->rbuf, 100);
     args->lines = (bcf1_t**) calloc(args->rbuf.m, sizeof(bcf1_t*));
@@ -1877,7 +1900,7 @@ static void init_data(args_t *args)
     }
     if ( args->atomize==SPLIT )
     {
-        args->abuf = abuf_init(args->hdr, SPLIT); 
+        args->abuf = abuf_init(args->hdr, SPLIT);
         abuf_set_opt(args->abuf, bcf_hdr_t*, BCF_HDR, args->out_hdr);
         if ( args->old_rec_tag )
             abuf_set_opt(args->abuf, const char*, INFO_TAG, args->old_rec_tag);
@@ -1987,7 +2010,9 @@ static bcf1_t *next_atomized_line(args_t *args)
 }
 static void normalize_vcf(args_t *args)
 {
-    args->out = hts_open(args->output_fname, hts_bcf_wmode2(args->output_type,args->output_fname));
+    char wmode[8];
+    set_wmode(wmode,args->output_type,args->output_fname,args->clevel);
+    args->out = hts_open(args->output_fname ? args->output_fname : "-", wmode);
     if ( args->out == NULL ) error("Can't write to \"%s\": %s\n", args->output_fname, strerror(errno));
     if ( args->n_threads )
         hts_set_opt(args->out, HTS_OPT_THREAD_POOL, args->files->p);
@@ -2081,16 +2106,19 @@ static void usage(void)
     fprintf(stderr, "        --force                     Try to proceed even if malformed tags are encountered. Experimental, use at your own risk\n");
     fprintf(stderr, "        --keep-sum TAG,..           Keep vector sum constant when splitting multiallelics (see github issue #360)\n");
     fprintf(stderr, "    -m, --multiallelics -|+TYPE     Split multiallelics (-) or join biallelics (+), type: snps|indels|both|any [both]\n");
+    fprintf(stderr, "        --multi-overlaps 0|.        Fill in the reference (0) or missing (.) allele when splitting multiallelics [0]\n");
     fprintf(stderr, "        --no-version                Do not append version and command line to the header\n");
     fprintf(stderr, "    -N, --do-not-normalize          Do not normalize indels (with -m or -c s)\n");
     fprintf(stderr, "        --old-rec-tag STR           Annotate modified records with INFO/STR indicating the original variant\n");
     fprintf(stderr, "    -o, --output FILE               Write output to a file [standard output]\n");
-    fprintf(stderr, "    -O, --output-type TYPE          'b' compressed BCF; 'u' uncompressed BCF; 'z' compressed VCF; 'v' uncompressed VCF [v]\n");
+    fprintf(stderr, "    -O, --output-type u|b|v|z[0-9]  u/b: un/compressed BCF, v/z: un/compressed VCF, 0-9: compression level [v]\n");
     fprintf(stderr, "    -r, --regions REGION            Restrict to comma-separated list of regions\n");
     fprintf(stderr, "    -R, --regions-file FILE         Restrict to regions listed in a file\n");
+    fprintf(stderr, "        --regions-overlap 0|1|2     Include if POS in the region (0), record overlaps (1), variant overlaps (2) [1]\n");
     fprintf(stderr, "    -s, --strict-filter             When merging (-m+), merged site is PASS only if all sites being merged PASS\n");
     fprintf(stderr, "    -t, --targets REGION            Similar to -r but streams rather than index-jumps\n");
     fprintf(stderr, "    -T, --targets-file FILE         Similar to -R but streams rather than index-jumps\n");
+    fprintf(stderr, "        --targets-overlap 0|1|2     Include if POS in the region (0), record overlaps (1), variant overlaps (2) [0]\n");
     fprintf(stderr, "        --threads INT               Use multithreading with <int> worker threads [0]\n");
     fprintf(stderr, "    -w, --site-win INT              Buffer for sorting lines which changed position during realignment [1000]\n");
     fprintf(stderr, "\n");
@@ -2118,9 +2146,13 @@ int main_vcfnorm(int argc, char *argv[])
     args->buf_win = 1000;
     args->mrows_collapse = COLLAPSE_BOTH;
     args->do_indels = 1;
+    args->ma_use_ref_allele = 1;
+    args->clevel = -1;
     int region_is_file  = 0;
     int targets_is_file = 0;
     args->use_star_allele = 1;
+    int regions_overlap = 1;
+    int targets_overlap = 0;
 
     static struct option loptions[] =
     {
@@ -2133,10 +2165,13 @@ int main_vcfnorm(int argc, char *argv[])
         {"fasta-ref",required_argument,NULL,'f'},
         {"do-not-normalize",no_argument,NULL,'N'},
         {"multiallelics",required_argument,NULL,'m'},
+        {"multi-overlaps",required_argument,NULL,13},
         {"regions",required_argument,NULL,'r'},
         {"regions-file",required_argument,NULL,'R'},
+        {"regions-overlap",required_argument,NULL,1},
         {"targets",required_argument,NULL,'t'},
         {"targets-file",required_argument,NULL,'T'},
+        {"targets-overlap",required_argument,NULL,2},
         {"site-win",required_argument,NULL,'w'},
         {"remove-duplicates",no_argument,NULL,'D'},
         {"rm-dup",required_argument,NULL,'d'},
@@ -2164,6 +2199,11 @@ int main_vcfnorm(int argc, char *argv[])
                 else error("Invalid argument to --atom-overlaps. Perhaps you wanted: \"--atom-overlaps '*'\"?\n");
                 break;
             case 12 : args->old_rec_tag = optarg; break;
+            case 13 :
+                if ( optarg[0]=='0' ) args->ma_use_ref_allele = 1;
+                else if ( optarg[0]=='.' ) args->ma_use_ref_allele = 0;
+                else error("Invalid argument to --multi-overlaps\n");
+                break;
             case 'N': args->do_indels = 0; break;
             case 'd':
                 if ( !strcmp("snps",optarg) ) args->rmdup = BCF_SR_PAIR_SNPS;
@@ -2200,12 +2240,21 @@ int main_vcfnorm(int argc, char *argv[])
                     case 'u': args->output_type = FT_BCF; break;
                     case 'z': args->output_type = FT_VCF_GZ; break;
                     case 'v': args->output_type = FT_VCF; break;
-                    default: error("The output type \"%s\" not recognised\n", optarg);
+                    default:
+                    {
+                        args->clevel = strtol(optarg,&tmp,10);
+                        if ( *tmp || args->clevel<0 || args->clevel>9 ) error("The output type \"%s\" not recognised\n", optarg);
+                    }
+                }
+                if ( optarg[1] )
+                {
+                    args->clevel = strtol(optarg+1,&tmp,10);
+                    if ( *tmp || args->clevel<0 || args->clevel>9 ) error("Could not parse argument: --compression-level %s\n", optarg+1);
                 }
                 break;
             case 'o': args->output_fname = optarg; break;
             case 'D':
-                fprintf(stderr,"Warning: `-D` is functional but deprecated, replaced by and alias of `-d none`.\n"); 
+                fprintf(stderr,"Warning: `-D` is functional but deprecated, replaced by and alias of `-d none`.\n");
                 args->rmdup = BCF_SR_PAIR_EXACT;
                 break;
             case 's': args->strict_filter = 1; break;
@@ -2221,6 +2270,14 @@ int main_vcfnorm(int argc, char *argv[])
             case  9 : args->n_threads = strtol(optarg, 0, 0); break;
             case  8 : args->record_cmd_line = 0; break;
             case  7 : args->force = 1; break;
+            case  1 :
+                regions_overlap = parse_overlap_option(optarg);
+                if ( regions_overlap < 0 ) error("Could not parse: --regions-overlap %s\n",optarg);
+                break;
+            case  2 :
+                targets_overlap = parse_overlap_option(optarg);
+                if ( targets_overlap < 0 ) error("Could not parse: --targets-overlap %s\n",optarg);
+                break;
             case 'h':
             case '?': usage(); break;
             default: error("Unknown argument: %s\n", optarg);
@@ -2241,11 +2298,13 @@ int main_vcfnorm(int argc, char *argv[])
 
     if ( args->region )
     {
+        bcf_sr_set_opt(args->files,BCF_SR_REGIONS_OVERLAP,regions_overlap);
         if ( bcf_sr_set_regions(args->files, args->region,region_is_file)<0 )
             error("Failed to read the regions: %s\n", args->region);
     }
     if ( args->targets )
     {
+        bcf_sr_set_opt(args->files,BCF_SR_TARGETS_OVERLAP,targets_overlap);
         if ( bcf_sr_set_targets(args->files, args->targets,targets_is_file, 0)<0 )
             error("Failed to read the targets: %s\n", args->targets);
     }
